#! /usr/bin/python3
"""This script checks that the symbols in a library match the given debian symbols, and
updates the debian symbols if needed.

USAGE: ./check-and-update-debian-symbols.py LIB_DIR_PATH LIB_NAME LIB_VERSION ABI_VERSION

To use: Go to your build folder and run "make check-miral-symbols"""

import sys
from sys import stderr
import os
from os import path
import subprocess
import re

HELPTEXT = __doc__

def get_output_symbols_path():
    return '/tmp/temp_mir_debian_symbols.symbols'

class Context:
    def __init__(self, args):
        assert isinstance(args, list)
        i = 0
        arg_len = 4
        if len(args) == 0 or '-h' in args or '--help' in args:
            raise RuntimeError('help')
        assert len(args) == arg_len, 'There should be exactly ' + str(arg_len) + ' arguments, not ' + str(len(args))
        self.library_dir_path = args[i]
        i += 1
        self.library_name = args[i]
        i += 1
        self.library_version = [int(i) for i in args[i].split('.')]
        i += 1
        self.abi_version = int(args[i])
        i += 1
        assert i == arg_len, 'i does not match arg_len'
        self.library_so_path = self.library_dir_path + '/lib' + self.library_name + '.so.' + str(self.abi_version)
        self.deb_package_name = 'lib' + self.library_name + str(self.abi_version)
        self.deb_symbols_path = 'debian/' + self.deb_package_name + '.symbols'
        self.output_symbols_path = get_output_symbols_path()

    def validate(self):
        assert path.isdir(self.library_dir_path), 'library dir path (' + self.library_dir_path + ') is not a valid directory'
        assert not '/' in self.library_name, 'library name should be a file name only, not a path'
        assert not self.library_name.startswith('lib'), 'library name should just be the name with no lib prefix'
        assert len(self.library_version) == 3, 'library version should be made of 3 dot separated components'
        assert path.isfile(self.library_so_path), 'library .so path (' + self.library_so_path + ') is not a valid file'
        assert path.isfile(self.deb_symbols_path), 'debian symbols file (' + self.deb_symbols_path + ') does not exist'

class Run:
    def __init__(self, args, cwd=None, stdin=None):
        assert isinstance(args, list)
        assert isinstance(cwd, str) or cwd == None
        assert isinstance(stdin, str) or stdin == None
        self.args = args
        # print('Running `' + ' '.join(self.args) + '`')
        p = subprocess.Popen(
                args,
                cwd=cwd,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
            )
        stdin = bytes(stdin, 'utf-8') if stdin != None else None
        stdout, stderr = p.communicate(stdin)
        self.stdout = stdout.decode('utf-8') if stdout != None else ''
        self.stderr = stderr.decode('utf-8') if stderr != None else ''
        self.returncode = p.returncode

    def success(self):
        return self.returncode == 0

    def assert_success(self):
        assert self.success(), (
            'Command `' + ' '.join(self.args) + '` failed with return code ' + str(self.returncode) + ':\n' +
            '  OUT: ' + '\n       '.join(line for line in self.stdout.split('\n')) + '\n' +
            '  ERR: ' + '\n       '.join(line for line in self.stderr.split('\n')))

class SymbolsFile:
    def __init__(self, context):
        assert isinstance(context, Context)
        self.context = context
        f = open(context.deb_symbols_path, 'r')
        lines = f.readlines()
        f.close()
        self.two_component_version = (
            str(context.library_version[0]) + '.' +
            str(context.library_version[1]))
        self.three_component_version = (
            str(context.library_version[0]) + '.' +
            str(context.library_version[1]) + '.' +
            str(context.library_version[2]))
        self.version_marker_line = (
            ' ' + context.library_name.upper() + '_' + self.two_component_version +
            '@' + context.library_name.upper() + '_' + self.two_component_version +
            ' ' + self.three_component_version + '\n')
        for i in reversed(range(len(lines))):
            if lines[i] == self.version_marker_line:
                self.prev_version_lines = lines[:i]
                self.current_version_lines = lines[i+1:]
                return
        self.prev_version_lines = lines
        self.current_version_lines = []

    def write(self):
        f = open(context.deb_symbols_path, 'w')
        for line in self.prev_version_lines:
            f.write(line)
        f.write(self.version_marker_line)
        for line in self.current_version_lines:
            f.write(line)
        f.close()

def new_lines_from_unfiltered(unfiltered):
    '''From the unfiltered (mangled) output of dpkg-gensymbols, returns a usable patch'''
    result = Run(['c++filt'], stdin=unfiltered)
    result.assert_success()
    searched = re.findall('^\+ (.+::.+) ([\.\d]+)$', result.stdout, flags=re.MULTILINE)
    if searched:
        filtered = [' (c++)"' + symbol + '" ' + version + '\n' for symbol, version in searched]
        return list(set(filtered))
    else:
        return None

def check_symbols(context):
    '''Checks the symbols.
    Returns None if all is good, or the lines that need to get added to the debian symbols file otherwise'''
    assert isinstance(context, Context)
    if (path.isfile(context.output_symbols_path)):
        os.remove(context.output_symbols_path)
    args = [
        'dpkg-gensymbols',
        '-e' + context.library_so_path,
        '-p' + context.deb_package_name,
        '-v' + '.'.join(str(i) for i in context.library_version),
        '-O' + context.output_symbols_path,
        '-c4',
    ]
    result = Run(args)
    if result.success():
        return None
    else:
        symbols_to_add = new_lines_from_unfiltered(result.stdout)
        if symbols_to_add != None:
            return symbols_to_add
        else:
            result.assert_success() # Something's not right. This will fail with a useful error

def append_to_symbols_file(new_symbols, context):
    '''Append a string to the end of the debian symbols file'''
    assert isinstance(new_symbols, list)
    symbols_file = SymbolsFile(context)
    symbols_file.current_version_lines += new_symbols
    symbols_file.current_version_lines.sort()
    symbols_file.write()

if __name__ == '__main__':
    context = Context(sys.argv[1:])
    context.validate()
    symbols_to_add = check_symbols(context)
    if symbols_to_add != None:
        print('Adding symbols to ' + context.deb_symbols_path + ':\n' + ''.join(symbols_to_add))
        append_to_symbols_file(symbols_to_add, context)
        print('Debian symbols added')



import subprocess
import babeltrace
import os
import tempfile
import atexit
from .common import unique_lttng_session_name

class LttngEvent:
    def __init__(self, bt_event):
        self.name = bt_event.name
        self.timestamp = bt_event.timestamp
        self.fields = {}
        for (k,v) in bt_event.items():
            self.fields[k] = v

    def __getitem__(self, k):
        return self.fields.get(k, None)

    def __str__(self):
        ts = str(self.timestamp)
        ts = ts[:-9] + "." + ts[-9:]
        return "[%s] %s %s" % (ts, self.name, self.fields)

class Lttng:
    def __init__(self, keep_lttng_traces):
        self.session_name = unique_lttng_session_name()
        self.output_dir = os.path.join(tempfile.gettempdir(), self.session_name)
        self.keep_lttng_traces = keep_lttng_traces

    def cleanup(self):
        subprocess.call("rm -rf %s" % self.output_dir, shell=True)

    def start(self):
        # Ensure we start with a clean session
        subprocess.call(["lttng", "stop"])
        subprocess.call(["lttng", "destroy", self.session_name])
        subprocess.call("rm -rf %s" % self.output_dir, shell=True)
        if not self.keep_lttng_traces:
            atexit.register(self.cleanup)
        subprocess.call(["lttng", "create", self.session_name, "-o", self.output_dir])
        subprocess.call(["lttng", "enable-event", "-u", "-a"])
        subprocess.call(["lttng", "add-context", "-u", "-t", "vpid"])
        subprocess.call(["lttng", "start"])

    def stop(self):
        subprocess.call(["lttng", "stop"])
        subprocess.call(["lttng", "destroy", self.session_name])

""" A performance test object """
class PerformanceTest:
    def __init__(self, components, keep_lttng_traces=False):
        """ Creates a PerformanceTest

        Args:
            components (list of Server and Client): An ordered list of test components to start.
            keep_lttng_traces (bool): Whether to keep the lttng trace directory produced by this run.
        """

        self.components = components
        self.lttng = Lttng(keep_lttng_traces)

    def start(self):
        """ Starts the performance test """
        self.lttng.start()
        for component in self.components:
            component.start()

    def stop(self):
        """ Stops the performance test"""
        for component in reversed(self.components):
            component.stop()
        self.lttng.stop()

    def babeltrace(self):
        """ Gets the babeltrace trace object for the test

        The test needs to have been stopped to get the trace.
        """
        col = babeltrace.TraceCollection()
        col.add_traces_recursive(self.lttng.output_dir, 'ctf')

        return col

    def events(self):
        """ Gets the trace events for the test

        The test needs to have been stopped to get the events.

        The events are returned in a list, making them easier to
        process than by using the raw babeltrace trace, which only
        supports iteration and access to one event at a time.
        """
        trace = self.babeltrace()

        events = []
        for e in trace.events:
            events.append(LttngEvent(e))

        return events



#! /usr/bin/python3
"""This script processes the XML generated by "make doc" and produces summary information
on symbols that libmiral intends to make public.

To use: Go to your build folder and run "make regenerate-miral-symbols-map" """

from xml.dom import minidom
from sys import argv

HELPTEXT = __doc__
DEBUG = False

def _get_text(node):
    substrings = []
    for node in node.childNodes:
        if node.nodeType == node.TEXT_NODE:
            substrings.append(node.data)
        elif node.nodeType == node.ELEMENT_NODE:
            substrings.append(_get_text(node))
    return ''.join(substrings)

def _get_text_for_element(parent, tagname):
    substrings = []

    for node in parent.getElementsByTagName(tagname):
        substrings.append(_get_text(node))

    return ''.join(substrings)

def _get_file_location(node):
    for node in node.childNodes:
        if node.nodeType == node.ELEMENT_NODE and node.tagName == 'location':
            return node.attributes['file'].value
    if DEBUG:
        print('no location in:', node)
    return None

def _has_element(node, tagname):
    for node in node.childNodes:
        if node.nodeType == node.ELEMENT_NODE and node.tagName in tagname:
            return True
    return False

def _print_attribs(node, attribs):
    for attrib in attribs:
        print(' ', attrib, '=', node.attributes[attrib].value)

def _concat_text_from_tags(parent, tagnames):
    substrings = []

    for tag in tagnames:
        substrings.append(_get_text_for_element(parent, tag))

    return ''.join(substrings)

def _print_location(node):
    print(' ', 'location', '=', _get_file_location(node))

def _get_attribs(node):
    kind = node.attributes['kind'].value
    static = node.attributes['static'].value
    prot = node.attributes['prot'].value
    return kind, static, prot

COMPONENT_MAP = {}
SYMBOLS = {'public' : set(), 'private' : set()}

def _report(publish, symbol):
    symbol = symbol.replace('~', '?')

    if publish:
        SYMBOLS['public'].add(symbol)
    else:
        SYMBOLS['private'].add(symbol)

    if not DEBUG:
        return

    if publish:
        print('  PUBLISH: {}'.format(symbol))
    else:
        print('NOPUBLISH: {}'.format(symbol))

OLD_STANZAS = '''MIRAL_3.0 {
global:
  extern "C++" {
    miral::AddInitCallback::?AddInitCallback*;
    miral::AddInitCallback::AddInitCallback*;
    miral::AddInitCallback::operator*;
    miral::AppendEventFilter::AppendEventFilter*;
    miral::AppendEventFilter::operator*;
    miral::ApplicationAuthorizer::?ApplicationAuthorizer*;
    miral::ApplicationAuthorizer::ApplicationAuthorizer*;
    miral::ApplicationAuthorizer::operator*;
    miral::ApplicationCredentials::ApplicationCredentials*;
    miral::ApplicationCredentials::gid*;
    miral::ApplicationCredentials::pid*;
    miral::ApplicationCredentials::uid*;
    miral::ApplicationInfo::?ApplicationInfo*;
    miral::ApplicationInfo::ApplicationInfo*;
    miral::ApplicationInfo::application*;
    miral::ApplicationInfo::name*;
    miral::ApplicationInfo::operator*;
    miral::ApplicationInfo::userdata*;
    miral::ApplicationInfo::windows*;
    miral::BasicSetApplicationAuthorizer::?BasicSetApplicationAuthorizer*;
    miral::BasicSetApplicationAuthorizer::BasicSetApplicationAuthorizer*;
    miral::BasicSetApplicationAuthorizer::operator*;
    miral::BasicSetApplicationAuthorizer::the_application_authorizer*;
    miral::CanonicalWindowManagerPolicy::CanonicalWindowManagerPolicy*;
    miral::CanonicalWindowManagerPolicy::advise_focus_gained*;
    miral::CanonicalWindowManagerPolicy::confirm_inherited_move*;
    miral::CanonicalWindowManagerPolicy::confirm_placement_on_display*;
    miral::CanonicalWindowManagerPolicy::handle_modify_window*;
    miral::CanonicalWindowManagerPolicy::handle_raise_window*;
    miral::CanonicalWindowManagerPolicy::handle_request_drag_and_drop*;
    miral::CanonicalWindowManagerPolicy::handle_window_ready*;
    miral::CanonicalWindowManagerPolicy::place_new_window*;
    miral::CommandLineOption::?CommandLineOption*;
    miral::CommandLineOption::CommandLineOption*;
    miral::CommandLineOption::operator*;
    miral::CursorTheme::?CursorTheme*;
    miral::CursorTheme::CursorTheme*;
    miral::CursorTheme::operator*;
    miral::DisplayConfiguration::?DisplayConfiguration*;
    miral::DisplayConfiguration::DisplayConfiguration*;
    miral::DisplayConfiguration::layout_option*;
    miral::DisplayConfiguration::list_layouts*;
    miral::DisplayConfiguration::operator*;
    miral::DisplayConfiguration::select_layout*;
    miral::ExternalClientLauncher::?ExternalClientLauncher*;
    miral::ExternalClientLauncher::ExternalClientLauncher*;
    miral::ExternalClientLauncher::launch*;
    miral::ExternalClientLauncher::launch_using_x11*;
    miral::ExternalClientLauncher::operator*;
    miral::ExternalClientLauncher::snapcraft_launch*;
    miral::InternalClientLauncher::?InternalClientLauncher*;
    miral::InternalClientLauncher::InternalClientLauncher*;
    miral::InternalClientLauncher::launch*;
    miral::InternalClientLauncher::operator*;
    miral::Keymap::?Keymap*;
    miral::Keymap::Keymap*;
    miral::Keymap::operator*;
    miral::Keymap::set_keymap*;
    miral::MinimalWindowManager::?MinimalWindowManager*;
    miral::MinimalWindowManager::MinimalWindowManager*;
    miral::MinimalWindowManager::advise_focus_gained*;
    miral::MinimalWindowManager::begin_pointer_move*;
    miral::MinimalWindowManager::begin_pointer_resize*;
    miral::MinimalWindowManager::begin_touch_move*;
    miral::MinimalWindowManager::begin_touch_resize*;
    miral::MinimalWindowManager::confirm_inherited_move*;
    miral::MinimalWindowManager::confirm_placement_on_display*;
    miral::MinimalWindowManager::handle_keyboard_event*;
    miral::MinimalWindowManager::handle_modify_window*;
    miral::MinimalWindowManager::handle_pointer_event*;
    miral::MinimalWindowManager::handle_raise_window*;
    miral::MinimalWindowManager::handle_request_drag_and_drop*;
    miral::MinimalWindowManager::handle_request_move*;
    miral::MinimalWindowManager::handle_request_resize*;
    miral::MinimalWindowManager::handle_touch_event*;
    miral::MinimalWindowManager::handle_window_ready*;
    miral::MinimalWindowManager::place_new_window*;
    miral::MirRunner::?MirRunner*;
    miral::MirRunner::MirRunner*;
    miral::MirRunner::add_start_callback*;
    miral::MirRunner::add_stop_callback*;
    miral::MirRunner::config_file*;
    miral::MirRunner::display_config_file*;
    miral::MirRunner::run_with*;
    miral::MirRunner::set_exception_handler*;
    miral::MirRunner::stop*;
    miral::MirRunner::wayland_display*;
    miral::MirRunner::x11_display*;
    miral::Output::?Output*;
    miral::Output::Output*;
    miral::Output::connected*;
    miral::Output::extents*;
    miral::Output::form_factor*;
    miral::Output::id*;
    miral::Output::is_same_output*;
    miral::Output::operator*;
    miral::Output::orientation*;
    miral::Output::physical_size_mm*;
    miral::Output::pixel_format*;
    miral::Output::power_mode*;
    miral::Output::refresh_rate*;
    miral::Output::scale*;
    miral::Output::type*;
    miral::Output::used*;
    miral::Output::valid*;
    miral::SetCommandLineHandler::?SetCommandLineHandler*;
    miral::SetCommandLineHandler::SetCommandLineHandler*;
    miral::SetCommandLineHandler::operator*;
    miral::SetTerminator::?SetTerminator*;
    miral::SetTerminator::SetTerminator*;
    miral::SetTerminator::operator*;
    miral::SetWindowManagementPolicy::?SetWindowManagementPolicy*;
    miral::SetWindowManagementPolicy::SetWindowManagementPolicy*;
    miral::SetWindowManagementPolicy::operator*;
    miral::StartupInternalClient::?StartupInternalClient*;
    miral::StartupInternalClient::StartupInternalClient*;
    miral::StartupInternalClient::operator*;
    miral::WaylandExtensions::?WaylandExtensions*;
#    miral::WaylandExtensions::Context::?Context*;
#    miral::WaylandExtensions::Context::Context*;
#    miral::WaylandExtensions::Context::operator*;
    miral::WaylandExtensions::WaylandExtensions*;
    miral::WaylandExtensions::add_extension*;
    miral::WaylandExtensions::add_extension_disabled_by_default*;
    miral::WaylandExtensions::all_supported*;
    miral::WaylandExtensions::disable*;
    miral::WaylandExtensions::enable*;
    miral::WaylandExtensions::operator*;
    miral::WaylandExtensions::recommended*;
    miral::WaylandExtensions::set_filter*;
    miral::WaylandExtensions::supported*;
    miral::WaylandExtensions::zwlr_layer_shell_v1*;
    miral::WaylandExtensions::zxdg_output_manager_v1*;
    miral::Window::?Window*;
    miral::Window::Window*;
    miral::Window::application*;
    miral::Window::move_to*;
    miral::Window::operator*;
    miral::Window::resize*;
    miral::Window::size*;
    miral::Window::top_left*;
    miral::WindowInfo::?WindowInfo*;
    miral::WindowInfo::WindowInfo*;
    miral::WindowInfo::application_id*;
    miral::WindowInfo::attached_edges*;
    miral::WindowInfo::can_be_active*;
    miral::WindowInfo::can_morph_to*;
    miral::WindowInfo::children*;
    miral::WindowInfo::clip_area*;
    miral::WindowInfo::confine_pointer*;
    miral::WindowInfo::constrain_resize*;
    miral::WindowInfo::depth_layer*;
    miral::WindowInfo::exclusive_rect*;
    miral::WindowInfo::has_output_id*;
    miral::WindowInfo::height_inc*;
    miral::WindowInfo::is_visible*;
    miral::WindowInfo::max_aspect*;
    miral::WindowInfo::max_height*;
    miral::WindowInfo::max_width*;
    miral::WindowInfo::min_aspect*;
    miral::WindowInfo::min_height*;
    miral::WindowInfo::min_width*;
    miral::WindowInfo::must_have_parent*;
    miral::WindowInfo::must_not_have_parent*;
    miral::WindowInfo::name*;
    miral::WindowInfo::needs_titlebar*;
    miral::WindowInfo::operator*;
    miral::WindowInfo::output_id*;
    miral::WindowInfo::parent*;
    miral::WindowInfo::preferred_orientation*;
    miral::WindowInfo::restore_rect*;
    miral::WindowInfo::shell_chrome*;
    miral::WindowInfo::state*;
    miral::WindowInfo::type*;
    miral::WindowInfo::userdata*;
    miral::WindowInfo::width_inc*;
    miral::WindowInfo::window*;
    miral::WindowManagementPolicy::?WindowManagementPolicy*;
    miral::WindowManagementPolicy::WindowManagementPolicy*;
    miral::WindowManagementPolicy::advise_adding_to_workspace*;
    miral::WindowManagementPolicy::advise_application_zone_create*;
    miral::WindowManagementPolicy::advise_application_zone_delete*;
    miral::WindowManagementPolicy::advise_application_zone_update*;
    miral::WindowManagementPolicy::advise_begin*;
    miral::WindowManagementPolicy::advise_delete_app*;
    miral::WindowManagementPolicy::advise_delete_window*;
    miral::WindowManagementPolicy::advise_end*;
    miral::WindowManagementPolicy::advise_focus_gained*;
    miral::WindowManagementPolicy::advise_focus_lost*;
    miral::WindowManagementPolicy::advise_move_to*;
    miral::WindowManagementPolicy::advise_new_app*;
    miral::WindowManagementPolicy::advise_new_window*;
    miral::WindowManagementPolicy::advise_output_create*;
    miral::WindowManagementPolicy::advise_output_delete*;
    miral::WindowManagementPolicy::advise_output_update*;
    miral::WindowManagementPolicy::advise_raise*;
    miral::WindowManagementPolicy::advise_removing_from_workspace*;
    miral::WindowManagementPolicy::advise_resize*;
    miral::WindowManagementPolicy::advise_state_change*;
    miral::WindowManagementPolicy::operator*;
    miral::WindowManagerOptions::WindowManagerOptions*;
    miral::WindowManagerOptions::operator*;
    miral::WindowManagerTools::?WindowManagerTools*;
    miral::WindowManagerTools::WindowManagerTools*;
    miral::WindowManagerTools::active_application_zone*;
    miral::WindowManagerTools::active_output*;
    miral::WindowManagerTools::active_window*;
    miral::WindowManagerTools::add_tree_to_workspace*;
    miral::WindowManagerTools::ask_client_to_close*;
    miral::WindowManagerTools::count_applications*;
    miral::WindowManagerTools::create_workspace*;
    miral::WindowManagerTools::drag_active_window*;
    miral::WindowManagerTools::drag_window*;
    miral::WindowManagerTools::end_drag_and_drop*;
    miral::WindowManagerTools::find_application*;
    miral::WindowManagerTools::focus_next_application*;
    miral::WindowManagerTools::focus_next_within_application*;
    miral::WindowManagerTools::focus_prev_application*;
    miral::WindowManagerTools::focus_prev_within_application*;
    miral::WindowManagerTools::for_each_application*;
    miral::WindowManagerTools::for_each_window_in_workspace*;
    miral::WindowManagerTools::for_each_workspace_containing*;
    miral::WindowManagerTools::id_for_window*;
    miral::WindowManagerTools::info_for*;
    miral::WindowManagerTools::info_for_window_id*;
    miral::WindowManagerTools::invoke_under_lock*;
    miral::WindowManagerTools::modify_window*;
    miral::WindowManagerTools::move_workspace_content_to_workspace*;
    miral::WindowManagerTools::operator*;
    miral::WindowManagerTools::place_and_size_for_state*;
    miral::WindowManagerTools::raise_tree*;
    miral::WindowManagerTools::remove_tree_from_workspace*;
    miral::WindowManagerTools::select_active_window*;
    miral::WindowManagerTools::start_drag_and_drop*;
    miral::WindowManagerTools::window_at*;
    miral::WindowSpecification::?WindowSpecification*;
    miral::WindowSpecification::WindowSpecification*;
    miral::WindowSpecification::application_id*;
    miral::WindowSpecification::attached_edges*;
    miral::WindowSpecification::aux_rect*;
    miral::WindowSpecification::aux_rect_placement_gravity*;
    miral::WindowSpecification::aux_rect_placement_offset*;
    miral::WindowSpecification::confine_pointer*;
    miral::WindowSpecification::depth_layer*;
    miral::WindowSpecification::exclusive_rect*;
    miral::WindowSpecification::height_inc*;
    miral::WindowSpecification::input_mode*;
    miral::WindowSpecification::input_shape*;
    miral::WindowSpecification::max_aspect*;
    miral::WindowSpecification::max_height*;
    miral::WindowSpecification::max_width*;
    miral::WindowSpecification::min_aspect*;
    miral::WindowSpecification::min_height*;
    miral::WindowSpecification::min_width*;
    miral::WindowSpecification::name*;
    miral::WindowSpecification::operator*;
    miral::WindowSpecification::output_id*;
    miral::WindowSpecification::parent*;
    miral::WindowSpecification::placement_hints*;
    miral::WindowSpecification::preferred_orientation*;
    miral::WindowSpecification::server_side_decorated*;
    miral::WindowSpecification::shell_chrome*;
    miral::WindowSpecification::size*;
    miral::WindowSpecification::state*;
    miral::WindowSpecification::top_left*;
    miral::WindowSpecification::type*;
    miral::WindowSpecification::update*;
    miral::WindowSpecification::userdata*;
    miral::WindowSpecification::width_inc*;
    miral::WindowSpecification::window_placement_gravity*;
    miral::X11Support::?X11Support*;
    miral::X11Support::X11Support*;
    miral::X11Support::operator*;
    miral::Zone::?Zone*;
    miral::Zone::Zone*;
    miral::Zone::extents*;
    miral::Zone::is_same_zone*;
    miral::Zone::operator*;
    miral::application_for*;
    miral::apply_lifecycle_state_to*;
    miral::display_configuration_options*;
    miral::equivalent_display_area*;
    miral::kill*;
    miral::name_of*;
    miral::operator*;
    miral::pid_of*;
    miral::pre_init*;
    miral::toolkit::mir_event_get_input_event*;
    miral::toolkit::mir_event_get_type*;
    miral::toolkit::mir_input_event_get_event*;
    miral::toolkit::mir_input_event_get_event_time*;
    miral::toolkit::mir_input_event_get_keyboard_event*;
    miral::toolkit::mir_input_event_get_pointer_event*;
    miral::toolkit::mir_input_event_get_touch_event*;
    miral::toolkit::mir_input_event_get_type*;
    miral::toolkit::mir_input_event_has_cookie*;
    miral::toolkit::mir_keyboard_event_action*;
    miral::toolkit::mir_keyboard_event_input_event*;
    miral::toolkit::mir_keyboard_event_key_code*;
    miral::toolkit::mir_keyboard_event_key_text*;
    miral::toolkit::mir_keyboard_event_modifiers*;
    miral::toolkit::mir_keyboard_event_scan_code*;
    miral::toolkit::mir_pointer_event_action*;
    miral::toolkit::mir_pointer_event_axis_value*;
    miral::toolkit::mir_pointer_event_button_state*;
    miral::toolkit::mir_pointer_event_buttons*;
    miral::toolkit::mir_pointer_event_input_event*;
    miral::toolkit::mir_pointer_event_modifiers*;
    miral::toolkit::mir_touch_event_action*;
    miral::toolkit::mir_touch_event_axis_value*;
    miral::toolkit::mir_touch_event_id*;
    miral::toolkit::mir_touch_event_input_event*;
    miral::toolkit::mir_touch_event_modifiers*;
    miral::toolkit::mir_touch_event_point_count*;
    miral::toolkit::mir_touch_event_tooltype*;
    miral::window_for*;
    non-virtual?thunk?to?miral::ApplicationAuthorizer::?ApplicationAuthorizer*;
    non-virtual?thunk?to?miral::CanonicalWindowManagerPolicy::advise_focus_gained*;
    non-virtual?thunk?to?miral::CanonicalWindowManagerPolicy::confirm_inherited_move*;
    non-virtual?thunk?to?miral::CanonicalWindowManagerPolicy::confirm_placement_on_display*;
    non-virtual?thunk?to?miral::CanonicalWindowManagerPolicy::handle_modify_window*;
    non-virtual?thunk?to?miral::CanonicalWindowManagerPolicy::handle_raise_window*;
    non-virtual?thunk?to?miral::CanonicalWindowManagerPolicy::handle_request_drag_and_drop*;
    non-virtual?thunk?to?miral::CanonicalWindowManagerPolicy::handle_window_ready*;
    non-virtual?thunk?to?miral::CanonicalWindowManagerPolicy::place_new_window*;
    non-virtual?thunk?to?miral::MinimalWindowManager::advise_focus_gained*;
    non-virtual?thunk?to?miral::MinimalWindowManager::confirm_inherited_move*;
    non-virtual?thunk?to?miral::MinimalWindowManager::confirm_placement_on_display*;
    non-virtual?thunk?to?miral::MinimalWindowManager::handle_keyboard_event*;
    non-virtual?thunk?to?miral::MinimalWindowManager::handle_modify_window*;
    non-virtual?thunk?to?miral::MinimalWindowManager::handle_pointer_event*;
    non-virtual?thunk?to?miral::MinimalWindowManager::handle_raise_window*;
    non-virtual?thunk?to?miral::MinimalWindowManager::handle_request_drag_and_drop*;
    non-virtual?thunk?to?miral::MinimalWindowManager::handle_request_move*;
    non-virtual?thunk?to?miral::MinimalWindowManager::handle_request_resize*;
    non-virtual?thunk?to?miral::MinimalWindowManager::handle_touch_event*;
    non-virtual?thunk?to?miral::MinimalWindowManager::handle_window_ready*;
    non-virtual?thunk?to?miral::MinimalWindowManager::place_new_window*;
#    non-virtual?thunk?to?miral::WaylandExtensions::Context::?Context*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::?WindowManagementPolicy*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_adding_to_workspace*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_application_zone_create*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_application_zone_delete*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_application_zone_update*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_begin*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_delete_app*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_delete_window*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_end*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_focus_gained*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_focus_lost*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_move_to*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_new_app*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_new_window*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_output_create*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_output_delete*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_output_update*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_raise*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_removing_from_workspace*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_resize*;
    non-virtual?thunk?to?miral::WindowManagementPolicy::advise_state_change*;
    typeinfo?for?miral::AddInitCallback;
    typeinfo?for?miral::AppendEventFilter;
    typeinfo?for?miral::ApplicationAuthorizer;
    typeinfo?for?miral::ApplicationCredentials;
    typeinfo?for?miral::ApplicationInfo;
    typeinfo?for?miral::BasicSetApplicationAuthorizer;
    typeinfo?for?miral::CanonicalWindowManagerPolicy;
    typeinfo?for?miral::CommandLineOption;
    typeinfo?for?miral::CursorTheme;
    typeinfo?for?miral::DisplayConfiguration;
    typeinfo?for?miral::ExternalClientLauncher;
    typeinfo?for?miral::InternalClientLauncher;
    typeinfo?for?miral::Keymap;
    typeinfo?for?miral::MinimalWindowManager;
    typeinfo?for?miral::MirRunner;
    typeinfo?for?miral::Output;
    typeinfo?for?miral::Output::PhysicalSizeMM;
    typeinfo?for?miral::SetCommandLineHandler;
    typeinfo?for?miral::SetTerminator;
    typeinfo?for?miral::SetWindowManagementPolicy;
    typeinfo?for?miral::StartupInternalClient;
    typeinfo?for?miral::WaylandExtensions;
    typeinfo?for?miral::WaylandExtensions::Builder;
    typeinfo?for?miral::WaylandExtensions::Context;
    typeinfo?for?miral::Window;
    typeinfo?for?miral::WindowInfo;
    typeinfo?for?miral::WindowManagementPolicy;
    typeinfo?for?miral::WindowManagerOption;
    typeinfo?for?miral::WindowManagerOptions;
    typeinfo?for?miral::WindowManagerTools;
    typeinfo?for?miral::WindowSpecification;
    typeinfo?for?miral::WindowSpecification::AspectRatio;
    typeinfo?for?miral::X11Support;
    typeinfo?for?miral::Zone;
    vtable?for?miral::AddInitCallback;
    vtable?for?miral::AppendEventFilter;
    vtable?for?miral::ApplicationAuthorizer;
    vtable?for?miral::ApplicationCredentials;
    vtable?for?miral::ApplicationInfo;
    vtable?for?miral::BasicSetApplicationAuthorizer;
    vtable?for?miral::CanonicalWindowManagerPolicy;
    vtable?for?miral::CommandLineOption;
    vtable?for?miral::CursorTheme;
    vtable?for?miral::DisplayConfiguration;
    vtable?for?miral::ExternalClientLauncher;
    vtable?for?miral::InternalClientLauncher;
    vtable?for?miral::Keymap;
    vtable?for?miral::MinimalWindowManager;
    vtable?for?miral::MirRunner;
    vtable?for?miral::Output;
    vtable?for?miral::Output::PhysicalSizeMM;
    vtable?for?miral::SetCommandLineHandler;
    vtable?for?miral::SetTerminator;
    vtable?for?miral::SetWindowManagementPolicy;
    vtable?for?miral::StartupInternalClient;
    vtable?for?miral::WaylandExtensions;
    vtable?for?miral::WaylandExtensions::Builder;
#    vtable?for?miral::WaylandExtensions::Context;
    vtable?for?miral::Window;
    vtable?for?miral::WindowInfo;
    vtable?for?miral::WindowManagementPolicy;
    vtable?for?miral::WindowManagerOption;
    vtable?for?miral::WindowManagerOptions;
    vtable?for?miral::WindowManagerTools;
    vtable?for?miral::WindowSpecification;
    vtable?for?miral::WindowSpecification::AspectRatio;
    vtable?for?miral::X11Support;
    vtable?for?miral::Zone;
  };
local: *;
};

MIRAL_3.1 {
global:
  extern "C++" {
    miral::WaylandExtensions::zwlr_foreign_toplevel_manager_v1*;
  };
} MIRAL_3.0;

MIRAL_3.2 {
global:'''

END_NEW_STANZA = '} MIRAL_3.1;'

def _print_report():
    print(OLD_STANZAS)
    new_symbols = False;
    for symbol in sorted(SYMBOLS['public']):
        formatted_symbol = '    {};'.format(symbol)
        if formatted_symbol not in OLD_STANZAS and 'miral::' in formatted_symbol:
            if not new_symbols:
                new_symbols = True;
                print('  extern "C++" {')
            print(formatted_symbol)

    if new_symbols: print("  };")
    print(END_NEW_STANZA)

def _print_debug_info(node, attributes):
    if not DEBUG:
        return
    print()
    _print_attribs(node, attributes)
    _print_location(node)

def _parse_member_def(context_name, node, is_class):
    kind = node.attributes['kind'].value

    if (kind in ['enum', 'typedef']
        or _has_element(node, ['templateparamlist'])
        or kind in ['function'] and node.attributes['inline'].value == 'yes'):
        return

    name = _concat_text_from_tags(node, ['name'])

    if name in ['__attribute__']:
        if DEBUG:
            print('  ignoring doxygen mis-parsing:', _concat_text_from_tags(node, ['argsstring']))
        return

    if name.startswith('operator'):
        name = 'operator'

    if not context_name is None:
        symbol = context_name + '::' + name
    else:
        symbol = name

    is_function = kind == 'function'

    if is_function:
        _print_debug_info(node, ['kind', 'prot', 'static', 'virt'])
    else:
        _print_debug_info(node, ['kind', 'prot', 'static'])

    if DEBUG:
        print('  is_class:', is_class)

    publish = _should_publish(is_class, is_function, node)

    _report(publish, symbol + '*')

    if is_function and node.attributes['virt'].value == 'virtual':
        _report(publish, 'non-virtual?thunk?to?' + symbol + '*')


def _should_publish(is_class, is_function, node):
    (kind, static, prot) = _get_attribs(node)

    publish = True

    if publish:
        publish = kind != 'define'

    if publish and is_class:
        publish = is_function or static == 'yes'

    if publish and prot == 'private':
        if is_function:
            publish = node.attributes['virt'].value == 'virtual'
        else:
            publish = False

    if publish and _has_element(node, ['argsstring']):
        publish = not _get_text_for_element(node, 'argsstring').endswith('=0')

    return publish


def _parse_compound_defs(xmldoc):
    compounddefs = xmldoc.getElementsByTagName('compounddef')
    for node in compounddefs:
        kind = node.attributes['kind'].value

        if kind in ['page', 'file', 'example', 'union']:
            continue

        if kind in ['group']:
            for member in node.getElementsByTagName('memberdef'):
                _parse_member_def(None, member, False)
            continue

        if kind in ['namespace']:
            symbol = _concat_text_from_tags(node, ['compoundname'])
            for member in node.getElementsByTagName('memberdef'):
                _parse_member_def(symbol, member, False)
            continue

        filename = _get_file_location(node)

        if DEBUG:
            print('  from file:', filename)

        if ('/examples/' in filename or '/test/' in filename or '[generated]' in filename
            or '[STL]' in filename or _has_element(node, ['templateparamlist'])):
            continue

        symbol = _concat_text_from_tags(node, ['compoundname'])

        publish = True

        if publish:
            if kind in ['class', 'struct']:
                prot = node.attributes['prot'].value
                publish = prot != 'private'
                _print_debug_info(node, ['kind', 'prot'])
                _report(publish, 'vtable?for?' + symbol)
                _report(publish, 'typeinfo?for?' + symbol)

        if publish:
            for member in node.getElementsByTagName('memberdef'):
                _parse_member_def(symbol, member, kind in ['class', 'struct'])

if __name__ == "__main__":
    if len(argv) == 1 or '-h' in argv or '--help' in argv:
        print(HELPTEXT)
        exit()

    for arg in argv[1:]:
        try:
            if DEBUG:
                print('Processing:', arg)
            _parse_compound_defs(minidom.parse(arg))
        except Exception as error:
            print('Error:', arg, error)

    if DEBUG:
        print('Processing complete')

    _print_report()



import os
import time
import subprocess
import shutil
from .common import env_variable_for_client_report, unique_server_socket_path

class Server:
    def __init__(self, executable=shutil.which("mir_demo_server"), host=None, reports=[], options=[], env={}):
        """ Creates a Server object.

        The server is not run until it is passed to a PerformanceTest and the
        test is started.

        Args:
            executable (str): The server executable to run.
            host (Server): The host server to connect to (or None if this is a host server)
            reports (list of str): The lttng reports to enable. Use the 'client-' prefix to
                                   enable a client report.
            options (list of str): Command-line options to use when running the server.
            env (dict of str=>str): Extra environment variables to set when running the server
        """
        self.executable = executable
        self.host = host
        self.reports = reports
        self.socket_path = unique_server_socket_path()
        self.options = options
        self.env = env
        self.process = None

    def start(self):
        cmdline = [self.executable, "-f", self.socket_path]
        if self.host is not None:
            cmdline.extend(["--host-socket", self.host.socket_path])

        env = os.environ.copy()
        env["LTTNG_UST_WITHOUT_BADDR_STATEDUMP"] = "1"
        for name,value in self.env.items():
            env[name] = value

        for report in self.reports:
            if report.startswith("client-"):
                env[env_variable_for_client_report(report)] = "lttng"
            else:
                cmdline.append("--%s-report=lttng" % report)

        cmdline.extend(self.options)

        self.process = subprocess.Popen(cmdline, env=env)
        self.wait_for_socket()

    def stop(self):
        self.process.terminate()

    def wait_for_socket(self):
        while not os.path.exists(self.socket_path):
            time.sleep(0.01)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2010 bit.ly
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Generate a text format histogram

This is a loose port to python of the Perl version at
http://www.pandamatak.com/people/anand/xfer/histo

http://github.com/bitly/data_hacks
"""
from decimal import Decimal
import math
import sys


class MVSD(object):
    """ A class that calculates a running Mean / Variance / Standard Deviation"""
    def __init__(self):
        self.is_started = False
        self.ss = Decimal(0) # (running) sum of square deviations from mean
        self.m = Decimal(0) # (running) mean
        self.total_w = Decimal(0) # weight of items seen

    def add(self, x, w=1):
        """ add another datapoint to the Mean / Variance / Standard Deviation"""
        if not isinstance(x, Decimal):
            x = Decimal(x)
        if not self.is_started:
            self.m = x
            self.ss = Decimal(0)
            self.total_w = w
            self.is_started = True
        else:
            temp_w = self.total_w + w
            self.ss += (self.total_w * w * (x - self.m) * (x - self.m )) / temp_w
            self.m += (x - self.m) / temp_w
            self.total_w = temp_w

        # print "added %-2d mean=%0.2f var=%0.2f std=%0.2f" % (x, self.mean(), self.var(), self.sd())

    def var(self):
        return self.ss / self.total_w

    def sd(self):
        return math.sqrt(self.var())

    def mean(self):
        return self.m

def test_mvsd():
    mvsd = MVSD()
    for x in range(10):
        mvsd.add(x)

    assert '%.2f' % mvsd.mean() == "4.50"
    assert '%.2f' % mvsd.var() == "8.25"
    assert '%.14f' % mvsd.sd() == "2.87228132326901"

def median(values):
    length = len(values)
    if length % 2:
        median_indeces = [length // 2]
    else:
        median_indeces = [length // 2-1, length // 2]

    values = sorted(values)
    return sum([values[i] for i in median_indeces]) / len(median_indeces)

def test_median():
    assert 6 == median([8, 7, 9, 1, 2, 6, 3]) # odd-sized list
    assert 4 == median([4, 5, 2, 1, 9, 10]) # even-sized int list. (4+5)/2 = 4
    assert "4.50" == "%.2f" % median([4.0, 5, 2, 1, 9, 10]) #even-sized float list. (4.0+5)/2 = 4.5


def histogram(stream, minimum=None, maximum=None, buckets=None, custbuckets=None, calc_msvd=True):
    """
    Loop over the stream and add each entry to the dataset, printing out at the end


    minimum: minimum value for graph
    maximum: maximum value for graph
    buckets: Number of buckets to use for the histogram
    custbuckets: Comma seperated list of bucket edges for the histogram
    calc_msvd: Calculate and display Mean, Variance and SD.
    """
    if not minimum or not maximum:
        # glob the iterator here so we can do min/max on it
        data = list(stream)
    else:
        data = stream
    bucket_scale = 1

    if minimum:
        min_v = Decimal(minimum)
    else:
        min_v = min(data)
    if maximum:
        max_v = Decimal(maximum)
    else:
        max_v = max(data)

    if not max_v > min_v:
        raise ValueError('max must be > min. max:%s min:%s' % (max_v, min_v))
    diff = max_v - min_v

    boundaries = []
    bucket_counts = []

    if custbuckets:
        bound = custbuckets.split(',')
        bound_sort = sorted(map(Decimal, bound))

        # if the last value is smaller than the maximum, replace it
        if bound_sort[-1] < max_v:
            bound_sort[-1] = max_v

        # iterate through the sorted list and append to boundaries
        for x in bound_sort:
            if x >= min_v and x <= max_v:
                boundaries.append(x)
            elif x >= max_v:
                boundaries.append(max_v)
                break

        # beware: the min_v is not included in the boundaries, so no need to do a -1!
        bucket_counts = [0 for x in range(len(boundaries))]
        buckets = len(boundaries)
    else:
        buckets = buckets or 10
        if buckets <= 0:
            raise ValueError('# of buckets must be > 0')
        step = diff / buckets
        bucket_counts = [0 for x in range(buckets)]
        for x in range(buckets):
            boundaries.append(min_v + (step * (x + 1)))

    skipped = 0
    samples = 0
    mvsd = MVSD()
    accepted_data = []
    for value in data:
        samples +=1
        if calc_msvd:
            mvsd.add(value)
            accepted_data.append(value)
        # find the bucket this goes in
        if value < min_v or value > max_v:
            skipped +=1
            continue
        for bucket_postion, boundary in enumerate(boundaries):
            if value <= boundary:
                bucket_counts[bucket_postion] +=1
                break

    # auto-pick the hash scale
    if max(bucket_counts) > 75:
        bucket_scale = int(max(bucket_counts) / 75)

    print("# NumSamples = %d; Min = %0.2f; Max = %0.2f" % (samples, min_v, max_v))
    if skipped:
        print("# %d value%s outside of min/max" % (skipped, skipped > 1 and 's' or ''))
    if calc_msvd:
        print("# Mean = %f; Variance = %f; SD = %f; Median %f" % (mvsd.mean(), mvsd.var(), mvsd.sd(), median(accepted_data)))
    print("# each ∎ represents a count of %d" % bucket_scale)
    bucket_min = min_v
    bucket_max = min_v
    for bucket in range(buckets):
        bucket_min = bucket_max
        bucket_max = boundaries[bucket]
        bucket_count = bucket_counts[bucket]
        star_count = 0
        if bucket_count:
            star_count = bucket_count // bucket_scale
        print('%10.4f - %10.4f [%6d]: %s' % (bucket_min, bucket_max, bucket_count, '∎' * star_count))




#!/usr/bin/env python3
# coding: utf-8

import json
import logging
import multiprocessing
import os
import pathlib
import pprint
import re
import sys
import subprocess
import tempfile

from launchpadlib import errors as lp_errors  # fades
from launchpadlib.credentials import (RequestTokenAuthorizationEngine,
                                      UnencryptedFileCredentialStore)
from launchpadlib.launchpad import Launchpad
import requests  # fades


logger = logging.getLogger("mir.process_snaps")
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)

APPLICATION = "mir-ci"
LAUNCHPAD = "production"
RELEASE = "focal"
TEAM = "mir-team"
SOURCE_NAME = "mir"
SNAPS = {
    "mir-kiosk": {
        "edge": {"ppa": "dev", "recipe": "mir-kiosk-edge"},
        "candidate": {"ppa": "rc", "recipe": "mir-kiosk-candidate"},
    },
    "mir-kiosk-apps": {
        "edge": {"recipe": "mir-kiosk-apps"},
    },
    "chromium-mir-kiosk": {
        "edge": {"recipe": "chromium-mir-kiosk"},
    },
    "mir-test-tools": {
        "edge": {"ppa": "dev", "recipe": "mir-test-tools-edge"},
        "candidate": {"ppa": "rc", "recipe": "mir-test-tools"},
    },
    "egmde": {
        "edge": {"ppa": "dev", "recipe": "egmde-mir-master"},
        "beta": {"ppa": "release", "recipe": "egmde-mir-release"},
    },
    "egmde-confined-desktop": {
        "edge": {"ppa": "dev", "recipe": "egmde-confined-desktop"},
    },
}


PENDING_BUILD = (
    "Needs building",
    "Dependency wait",
    "Currently building",
    "Uploading build",
)


IGNORED_SOURCES = (
    "Deleted",
    "Obsolete",
)

# See https://regex101.com/r/b8otIx/2
MIR_VERSION_RE = re.compile(r"^(?P<version>[0-9\.]+)"              # major.minor.patch
                            r"(?:(?P<build>[+~](?:rc|dev)[0-9]+)"  # optional [~+]{rc,dev}* build tag
                            r"-g(?P<commit>[0-9a-f]+))?"           # ... with git suffix
                            r"-(?P<distro>[^-]+)$")                # distro

SNAP_VERSION_RE = re.compile(r"^(?:(?P<server>.+)-mir)?"
                             r"(?P<mir>.+?)"
                             r"(?:-snap(?P<snap>.+))?$")

STORE_URL = ("https://api.snapcraft.io/api/v1/snaps"
             "/details/{snap}?channel={channel}")
STORE_HEADERS = {
    "X-Ubuntu-Series": "16",
    "X-Ubuntu-Architecture": "{arch}"
}

CHECK_NOTICES_PATH = "/snap/bin/review-tools.check-notices"


def get_store_snap(processor, snap, channel):
    logger.debug("Checking for snap %s on %s in channel %s", snap, processor, channel)
    data = {
        "snap": snap,
        "channel": channel,
        "arch": processor,
    }
    resp = requests.get(
        STORE_URL.format(**data),
        headers={k: v.format(**data) for k, v in STORE_HEADERS.items()}
    )
    logger.debug("Got store response: %s", resp)

    try:
        result = json.loads(resp.content)
    except json.JSONDecodeError:
        logger.error("Could not parse store response: %s",
                     resp.content)
    else:
        return result


def fetch_url(entry):
    dest, uri = entry
    r = requests.get(uri, stream=True)
    logger.debug("Downloading %s to %s…", uri, dest)
    if r.status_code == 200:
        with open(dest, "wb") as f:
            for chunk in r:
                f.write(chunk)
    return dest


def check_snap_notices(store_snaps):
    with tempfile.TemporaryDirectory(dir=pathlib.Path.home()) as dir:
        snaps = multiprocessing.Pool(8).map(
            fetch_url,
            ((pathlib.Path(dir) / f"{snap['package_name']}_{snap['revision']}.snap",
                snap["download_url"])
                for snap in store_snaps)
        )

        try:
            notices = subprocess.check_output([CHECK_NOTICES_PATH] + snaps)
            logger.debug("Got check_notices output:\n%s", notices.decode())
        except subprocess.CalledProcessError as e:
            logger.error("Failed to check notices:\n%s", e.output)
        else:
            notices = json.loads(notices)
            return notices


if __name__ == '__main__':
    try:
        lp = Launchpad.login_with(
            APPLICATION,
            LAUNCHPAD,
            version="devel",
            authorization_engine=RequestTokenAuthorizationEngine(LAUNCHPAD,
                                                                 APPLICATION),
            credential_store=UnencryptedFileCredentialStore(os.path.expanduser(sys.argv[1])),
        )
    except NotImplementedError:
        raise RuntimeError("Invalid credentials.")

    check_notices = (os.path.isfile(CHECK_NOTICES_PATH)
                     and os.access(CHECK_NOTICES_PATH, os.X_OK)
                     and CHECK_NOTICES_PATH)
    if not check_notices:
        logger.info("`review-tools` not found, skipping USN checks…")

    ubuntu = lp.distributions["ubuntu"]
    logger.debug("Got ubuntu: %s", ubuntu)

    series = ubuntu.getSeries(name_or_version=RELEASE)
    logger.debug("Got series: %s", series)

    team = lp.people[TEAM]
    logger.debug("Got team: %s", team)

    errors = []

    for snap, channels in SNAPS.items():
        for channel, snap_map in channels.items():
            logger.info("Processing channel %s for snap %s…", channel, snap)

            try:
                snap_recipe = lp.snaps.getByName(owner=team, name=snap_map["recipe"])
                logger.debug("Got snap: %s", snap_recipe)
            except lp_errors.NotFound as ex:
                logger.error("Snap not found: %s", snap_map["recipe"])
                errors.append(ex)
                continue

            if "ppa" in snap_map:
                ppa = team.getPPAByName(name=snap_map["ppa"])
                latest_source = None
                logger.debug("Got ppa: %s", ppa)

                sources = ppa.getPublishedSources(
                    source_name=SOURCE_NAME,
                    distro_series=series)

                if not sources:
                    logger.error("Did not find %s in %s/%s", SOURCE_NAME, TEAM, snap_map["ppa"])
                    continue

                for source in sources:
                    if source.status in IGNORED_SOURCES:
                        logger.debug("Ignoring source: %s (status: %s)", source.display_name, source.status)
                        continue
                    latest_source = source
                    break

                if not latest_source:
                    logger.error("Did not find a valid source for %s in %s/%s", SOURCE_NAME, TEAM, snap_map["ppa"])
                    continue

                logger.debug("Latest source: %s", latest_source.display_name)

                mir_version = (
                    "".join(v or "" for v in MIR_VERSION_RE.fullmatch(latest_source.source_package_version).groups()[0:2])
                )
                logger.debug("Parsed upstream version: %s", mir_version)

                if latest_source.status != "Published":
                    logger.info("Skipping %s: %s",
                                latest_source.display_name, latest_source.status)
                    continue

                if any(build.buildstate in PENDING_BUILD
                    for build in latest_source.getBuilds()):
                    logger.info("Skipping %s: builds pending…",
                                latest_source.display_name)
                    continue

                if any(binary.status != "Published"
                    for binary in latest_source.getPublishedBinaries()):
                    logger.info("Skipping %s: binaries pending…",
                                latest_source.display_name)
                    continue
            else:
                mir_version = None

            if len(snap_recipe.pending_builds) > 0:
                logger.info("Skipping %s: snap builds pending…",
                            snap_recipe.web_link)
                continue

            store_snaps = tuple(filter(lambda snap: snap.get("result") != "error", (
                get_store_snap(processor.name,
                               snap,
                               channel)
                for processor in snap_recipe.processors
            )))

            logger.debug("Got store versions: %s", {snap["architecture"][0]: snap["version"] for snap in store_snaps})

            if all(store_snap is None
                   or mir_version is None
                   or mir_version == SNAP_VERSION_RE.match(store_snap["version"]).group("mir")
                   for store_snap in store_snaps):

                if check_notices:
                    snap_notices = check_snap_notices(store_snaps)[snap]

                    if any(snap_notices.values()):
                        logger.info("Found USNs:\n%s", pprint.pformat(snap_notices))
                    else:
                        logger.info(
                            "Skipping %s: store versions are current and no USNs found",
                            snap
                        )
                        continue

                else:
                    logger.info(
                        "Skipping %s: store versions are current",
                        snap
                    )
                    continue

            logger.info("Triggering %s…", snap_recipe.description or snap_recipe.name)

            snap_recipe.requestBuilds(archive=snap_recipe.auto_build_archive,
                                      pocket=snap_recipe.auto_build_pocket,
                                      channels=snap_recipe.auto_build_channels)
            logger.debug("Triggered builds: %s", snap_recipe.web_link)

    for error in errors:
        logger.debug(error)

    if errors:
        sys.exit(1)



#!/usr/bin/python3

from mir_perf_framework import PerformanceTest, Server, Client
import time
import evdev
import statistics
import subprocess

###### Helper classes ######

class TouchScreen:
    def __init__(self):
        res = self.get_resolution()

        allowed_events = {
            evdev.ecodes.EV_ABS : (
                (evdev.ecodes.ABS_MT_POSITION_X, (0, res[0], 0, 0)),
                (evdev.ecodes.ABS_MT_POSITION_Y, (0, res[1], 0, 0)),
                (evdev.ecodes.ABS_MT_TOUCH_MAJOR, (0, 30, 0, 0)),
                (evdev.ecodes.ABS_MT_TRACKING_ID, (0, 65535, 0, 0)),
                (evdev.ecodes.ABS_MT_PRESSURE, (0, 255, 0, 0)),
                (evdev.ecodes.ABS_MT_SLOT, (0, 9, 0, 0))
                ),
            evdev.ecodes.EV_KEY: [
                evdev.ecodes.BTN_TOUCH,
                ]
        }
        
        self.ui = evdev.UInput(events=allowed_events, name="autopilot-finger")

    def get_resolution(self):
        out = subprocess.check_output(["fbset", "-s"])
        out_list = out.split()
        geometry = out_list.index(b"geometry")
        return (int(out_list[geometry + 1]), int(out_list[geometry + 2]))

    def finger_down_at(self, xy):
        self.ui.write(evdev.ecodes.EV_ABS, evdev.ecodes.ABS_MT_SLOT, 0)
        self.ui.write(evdev.ecodes.EV_ABS, evdev.ecodes.ABS_MT_TRACKING_ID, 0)
        self.ui.write(evdev.ecodes.EV_KEY, evdev.ecodes.BTN_TOUCH, 1)
        self.ui.write(evdev.ecodes.EV_ABS, evdev.ecodes.ABS_MT_POSITION_X, xy[0])
        self.ui.write(evdev.ecodes.EV_ABS, evdev.ecodes.ABS_MT_POSITION_Y, xy[1])
        self.ui.write(evdev.ecodes.EV_ABS, evdev.ecodes.ABS_MT_PRESSURE, 50)
        self.ui.write(evdev.ecodes.EV_ABS, evdev.ecodes.ABS_MT_TOUCH_MAJOR, 4)
        self.ui.syn()

    def finger_up(self):
        self.ui.write(evdev.ecodes.EV_ABS, evdev.ecodes.ABS_MT_TRACKING_ID, -1)
        self.ui.syn()

    def finger_move_to(self, xy):
        self.ui.write(evdev.ecodes.EV_ABS, evdev.ecodes.ABS_MT_SLOT, 0)
        self.ui.write(evdev.ecodes.EV_ABS, evdev.ecodes.ABS_MT_POSITION_X, xy[0])
        self.ui.write(evdev.ecodes.EV_ABS, evdev.ecodes.ABS_MT_POSITION_Y, xy[1])
        self.ui.syn()


####### TEST #######

host = Server(reports=["input"])
nested = Server(host=host, reports=["client-input-receiver"])
client = Client(server=nested, reports=["client-input-receiver"], options=["-f"])

test = PerformanceTest([host, nested, client])
touch_screen = TouchScreen()

test.start()

# Perform three 1-second drag movements
for i in range(3):
    xy = (100,100)
    touch_screen.finger_down_at(xy)
    for i in range(100):
        touch_screen.finger_move_to(xy)
        xy = (xy[0] + 5, xy[1] + 5)
        time.sleep(0.01)
    touch_screen.finger_up()

test.stop()

####### TRACE PARSING #######

trace = test.babeltrace()

pids = {}
data = {}

for event in trace.events:
    if event.name == "mir_client_input_receiver:touch_event":
        pid = event["vpid"]
        if pid not in pids.values():
            if "nested" not in pids:
                pids["nested"] = pid
            elif "client" not in pids:
                pids["client"] = pid
        if pid not in data: data[pid] = []

        data[pid].append((event.timestamp - event["event_time"]) / 1000000.0)


print("=== Results ===")

nested_data = data[pids["nested"]]
print("Nested server received %d events" % len(nested_data))
print("Kernel to nested server mean: %f ms stdev: %f ms" %
      (statistics.mean(nested_data), statistics.stdev(nested_data)))

client_data = data[pids["client"]]
print("Client received %d events" % len(client_data))
print("Kernel to client mean: %f ms stdev: %f ms" %
      (statistics.mean(client_data), statistics.stdev(client_data)))



#!/usr/bin/env python3

import babeltrace
import sys
from text_histogram import histogram

howto = """
Process a LTTNG trace for per-client commit-to-present latencies.

This works on an lttng trace with the mir_server_wayland:sw_buffer_committed
and mir_server_compositor:buffers_in_frame tracepoints enabled. To generate
such a trace you need to run a Mir server with the --compositor-report=lttng
option. For example:

> miral-app --compositor-report=lttng

Once you have your compositor active, you can get an LTTNG trace like so:
> lttng create mir-trace
> lttng enable-event -u "mir_server_wayland:sw_buffer_committed"
> lttng enable-event -u "mir_server_compositor:buffers_in_frame"
> lttng start

At this point you can interact with applications in your Mir session;
LTTNG will be recording the necessary events. When you're done…

> lttng stop
> lttng destroy

This will (by default) leave you with a trace in
~/lttng-traces/mir-trace-DIGITS-MORE_DIGITS/ust/uid/$YOUR_UID/64-bit

This trace can then be processed with
> python3 commit_to_present.py ~/lttng-traces/mir-trace-$DIGITS-$MORE_DIGITS/ust/uid/$YOUR_UID/64-bit
"""

try:
    trace_path = sys.argv[1]

    trace_collection = babeltrace.TraceCollection()
    trace_collection.add_trace(trace_path, 'ctf')
except:
    print(howto)
    raise

class ClientStats:
    def __init__(self):
        self.buffers_in_flight = dict()
        self.buffer_delays = []

clients = dict()
for event in trace_collection.events:
    if event.name == 'mir_server_wayland:sw_buffer_committed':
        if event['client'] not in clients:
            clients[event['client']] = ClientStats()
        clients[event['client']].buffers_in_flight[event['buffer_id']] = event.timestamp
    if event.name == 'mir_server_compositor:buffers_in_frame':
        for buffer in event['buffer_ids']:
            for client in clients.values():
                if buffer in client.buffers_in_flight.keys():
                    client.buffer_delays.append(event.timestamp - client.buffers_in_flight[buffer])
                    del client.buffers_in_flight[buffer]

for (id, client)  in clients.items():
    print("Client: ", id)
    histogram([delay // 1000000 for delay in client.buffer_delays])



#!/usr/bin/python3

from mir_perf_framework import PerformanceTest, Server, Client
import time
import evdev
import statistics

####### TEST #######

host = Server()
nested = Server(host=host, reports=["client-input-receiver"])
client = Client(server=nested, reports=["client-input-receiver"])

ui = evdev.UInput()
test = PerformanceTest([host, nested, client])

test.start()

for i in range(1000):
    ui.write(evdev.ecodes.EV_KEY, evdev.ecodes.KEY_A, 1)
    ui.syn()
    time.sleep(0.002)
    ui.write(evdev.ecodes.EV_KEY, evdev.ecodes.KEY_A, 0)
    ui.syn()
    time.sleep(0.002)

test.stop()

####### TRACE PARSING #######

trace = test.babeltrace()

pids = {}
data = {}

for event in trace.events:
    if event.name == "mir_client_input_receiver:key_event":
        pid = event["vpid"]
        if pid not in pids.values():
            if "nested" not in pids:
                pids["nested"] = pid
            elif "client" not in pids:
                pids["client"] = pid
        if pid not in data: data[pid] = []

        data[pid].append((event.timestamp - event["event_time"]) / 1000000.0)


print("=== Results ===")

nested_data = data[pids["nested"]]
print("Kernel to nested mean: %f ms stdev: %f ms" %
      (statistics.mean(nested_data), statistics.stdev(nested_data)))

client_data = data[pids["client"]]
print("Kernel to client mean: %f ms stdev: %f ms" %
      (statistics.mean(client_data), statistics.stdev(client_data)))



import os
import subprocess
import shutil
import time
from .common import env_variable_for_client_report

class Client:
    def __init__(self, executable=shutil.which("mir_demo_client_egltriangle"), server=None, reports=[], options=[], env={}):
        """ Creates a Client object.

        The client is not run until it is passed to a PerformanceTest and the
        test is started.

        Args:
            executable (str): The client executable to run.
            server (Server): The server to connect to.
            reports (list of str): The lttng reports to enable. Use the 'client-' prefix to
                                   enable a client report.
            options (list of str): Command-line options to use when running the client.
            env (dict of str=>str): Extra environment variables to set when running the client
        """
        self.executable = executable
        self.server = server
        self.reports = reports
        self.options = options
        self.env = env
        self.process = None

    def start(self):
        cmdline = [self.executable, "-m", self.server.socket_path]
        cmdline.extend(self.options)

        env = os.environ.copy()
        env["LTTNG_UST_WITHOUT_BADDR_STATEDUMP"] = "1"
        for report in self.reports:
            env[env_variable_for_client_report(report)] = "lttng"
        for name,value in self.env.items():
            env[name] = value

        self.process = subprocess.Popen(cmdline, env=env)
        # Wait for client to start
        # TODO: Find a better way to wait for client readiness
        time.sleep(0.25)

    def stop(self):
        self.process.terminate()



import tempfile
import os
import random
import string

random_str = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(6))

def env_variable_for_client_report(i):
    return "MIR_%s_REPORT" % i.upper().replace("-", "_")

def unique_server_socket_path():
    unique_server_socket_path.server_count += 1
    tmpname = "mir_perf_socket_" + str(os.getppid()) + "_"
    tmpname += str(random_str) + "_"
    tmpname += str(unique_server_socket_path.server_count)
    return os.path.join(tempfile.gettempdir(), tmpname)

unique_server_socket_path.server_count = 0

def unique_lttng_session_name():
    return "mir_perf_lttng_" + str(os.getpid()) + "_" + random_str



#!/usr/bin/env python
# coding: utf-8

# Copyright © 2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 or 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Authored by: Alexandros Frantzis <alexandros.frantzis@canonical.com>

import sys
from PIL import Image

def premultiply(image):
    pixels = image.load()
    for i in range(image.size[0]):
        for j in range(image.size[1]):
            orig = pixels[i,j]
            m = orig[3] / 255.0
            pixels[i,j] = (int(orig[0] * m) , int(orig[1] * m), int(orig[2] * m), orig[3])

def tocstring(data):
    result = ''
    line_chars = 0
    line_limit = 80

    for c in data:
        if line_chars == 0:
            result += '    "'

        s = '\\%o' % c
        result += s
        line_chars += len(s)

        if line_chars >= line_limit:
            result += '"\n'
            line_chars = 0

    if line_chars != 0:
        result += '"'

    return result

def bytes_per_pixel(image):
    if image.mode == 'RGBA':
        return 4
    elif image.mode == 'RGB':
        return 3
    else:
        raise "Unsupported image mode %s" % image.mode

def export(image, variable_name):
    image_info = (image.size[0], image.size[1], bytes_per_pixel(image))
    print("static const struct {")
    print("    unsigned int width;")
    print("    unsigned int height;")
    print("    unsigned int bytes_per_pixel; /* 3:RGB, 4:RGBA */")
    print("    unsigned char pixel_data[%d * %d * %d + 1];" % image_info)
    print("} %s = {" % variable_name)
    print("    %d, %d, %d," % image_info)
    print(tocstring(image.tobytes()))
    print("};")

def show_usage():
    print("Usage: ./png2header.py PNGFILE VARNAME > HEADER_FILE", file=sys.stderr)
    print("Convert a PNG image to an embeddable C/C++ header file", file=sys.stderr)

if len(sys.argv) < 3:
    show_usage()
    sys.exit(1)

image_filename = sys.argv[1]
variable_name = sys.argv[2]

image = Image.open(image_filename)
if image.mode == 'RGBA':
    premultiply(image)
export(image, variable_name)



#!/usr/bin/python3

from mir_perf_framework import PerformanceTest, Server, Client
import time
import statistics
import itertools

####### TEST #######

host = Server(reports=["display", "compositor"])
nested = Server(host=host, reports=["client-perf", "compositor"])
client = Client(server=nested, reports=["client-perf"])

test = PerformanceTest([host, nested, client])

test.start()

time.sleep(5)

test.stop()

####### TRACE PARSING #######

class Frame:
    def __init__(self, events, pids, event_index):
        self.events = events
        self.pids = pids
        self.event_index = event_index

    def find_frame_containing_this_one(self, pid, frame_end_event):
        frame = None
        buffer_in_frame = False

        for i, event in enumerate(itertools.islice(self.events, self.event_index, len(self.events)), start=self.event_index):
            if event["vpid"] != pid: continue
            if event.name == frame_end_event and buffer_in_frame:
                frame = Frame(self.events, self.pids, i)
                break
            if event.name == "mir_server_compositor:buffers_in_frame" and self.buffer_id() in event["buffer_ids"]:
                buffer_in_frame = True

        return frame

    def find_nested_frame_containing_this_one(self):
        return self.find_frame_containing_this_one(pids["nested"], "mir_client_perf:end_frame")

    def find_host_frame_containing_this_one(self):
        return self.find_frame_containing_this_one(pids["host"], "mir_server_display:report_vsync")

    def buffer_id(self):
        return self.events[self.event_index]["buffer_id"]

    def timestamp(self):
        return self.events[self.event_index].timestamp

def find_pids(events):
    pids = {}

    for event in events:
        if event.name == "mir_server_compositor:started":
            pid = event["vpid"]
            if "host" not in pids:
                pids["host"] = pid
            else:
                pids["nested"] = pid
                break

    for event in events:
        pid = event["vpid"]
        if pid not in pids.values():
            pids["client"] = pid
            break

    return pids

def find_client_frame_events(events, pids):
    client_frames = []
    for i, event in enumerate(events):
        if event["vpid"] == pids["client"] and event.name == "mir_client_perf:end_frame":
            client_frames.append(Frame(events, pids, i))
    return client_frames

events = test.events()
pids = find_pids(events)
client_frame_events = find_client_frame_events(events, pids)
data = []

for client_frame in client_frame_events:
    nested_frame = client_frame.find_nested_frame_containing_this_one()
    if nested_frame is None: continue

    host_frame = nested_frame.find_host_frame_containing_this_one()
    if host_frame is None: continue

    data.append((host_frame.timestamp() - client_frame.timestamp()) / 1000000.0)

print("=== Results ===")
print("Tracked %d buffers from nested client to display" % len(data))
print("Latency mean: %f ms stdev: %f ms" %
      (statistics.mean(data), statistics.stdev(data)))



#! /usr/bin/python
from xml.dom import minidom
from sys import argv

helptext = \
"""This script processes the XML generated by "make doc" and produces summary information
on symbols that Mir intends to make public.

To use:

1. Go to your build folder and run "make doc"
2. Create summary by running "../tools/process_doxygen_xml.py doc/xml/*.xml > raw-output"
3. For each symbol map (e.g. mirserver)
3.1. cat raw-output | grep "^mirserver public" | sed "s/mirserver public: /    /" | sort > mirserver-public-symbols
3.2. vi -d mirserver-public-symbols ../src/server/symbols.map"""

debug = False

def get_text(node):
    rc = []
    for node in node.childNodes:
        if node.nodeType == node.TEXT_NODE:
            rc.append(node.data)
        elif node.nodeType == node.ELEMENT_NODE:
            rc.append(get_text(node))
    return ''.join(rc)

def get_text_for_element(parent, tagname):
    rc = []
    nodes = parent.getElementsByTagName(tagname);
    for node in nodes : rc.append(get_text(node))
    return ''.join(rc)

def get_file_location(node):
    for node in node.childNodes:
        if node.nodeType == node.ELEMENT_NODE and node.tagName == 'location':
            return node.attributes['file'].value
    if debug: print 'no location in:', node
    return None

def has_element(node, tagname):
    for node in node.childNodes:
        if node.nodeType == node.ELEMENT_NODE and node.tagName in tagname:
            return True
    return False

def print_attribs(node, attribs):
    for attrib in attribs : print ' ', attrib, '=', node.attributes[attrib].value

def concat_text_from_tags(parent, tagnames):
    rc = []
    for tag in tagnames : rc.append(get_text_for_element(parent, tag))
    return ''.join(rc)

def print_location(node):
    print ' ', 'location', '=', get_file_location(node)

def get_attribs(node):
    kind = node.attributes['kind'].value
    static = node.attributes['static'].value
    prot =  node.attributes['prot'].value
    return (kind, static, prot)

def is_file_publishable(file_location):
    return file_location.startswith('include/') \
           or 'build/src' in file_location

# Special cases for publishing anyway:
publish_special_cases = {
    # Although private this is called by a template wrapper function that instantiates
    # in client code
    'mir::SharedLibrary::load_symbol*',
}

component_map = {}

def report(component, publish, symbol):
    symbol = symbol.replace('~', '?')

    if symbol in publish_special_cases: publish = True

    symbols = component_map.get(component, {'public' : set(), 'private' : set()})
    if publish: symbols['public'].add(symbol)
    else:       symbols['private'].add(symbol)
    component_map[component] = symbols
    if not debug: return
    if publish: print '  PUBLISH in {}: {}'.format(component, symbol)
    else      : print 'NOPUBLISH in {}: {}'.format(component, symbol)

def print_report():
    format = '{} {}: {};'
    for component, symbols in component_map.iteritems():
        print 'COMPONENT:', component
        for key in symbols.keys():
            for symbol in symbols[key]: print format.format(component, key, symbol)
        print

def print_debug_info(node, attributes):
    if not debug: return
    print
    print_attribs(node, attributes)
    print_location(node)

def find_physical_component(location_file):
    path_elements = location_file.split('/')
    found = False
    for element in path_elements:
        if found: return element
        found = element in ['include', 'src']
    if debug: print 'no component in:', location_file
    return None

def mapped_physical_component(location_file):
    location = find_physical_component(location_file)
    return 'mir' + location

def parse_member_def(context_name, node, is_class):
    library = mapped_physical_component(get_file_location(node))
    (kind, static, prot) = get_attribs(node)

    if kind in ['enum', 'typedef']: return
    if has_element(node, ['templateparamlist']): return
    if kind in ['function'] and node.attributes['inline'].value == 'yes': return

    name = concat_text_from_tags(node, ['name'])
    if name in ['__attribute__']:
        if debug: print '  ignoring doxygen mis-parsing:', concat_text_from_tags(node, ['argsstring'])
        return

    if name.startswith('operator'): name = 'operator'
    if not context_name == None: symbol = context_name + '::' + name
    else: symbol = name

    file_location = get_file_location(node)
    publish = is_file_publishable(file_location)

    is_function = kind == 'function'
    if publish: publish = kind != 'define'
    if publish and is_class: publish = is_function or static == 'yes'
    if publish and prot == 'private':
        if is_function: publish = node.attributes['virt'].value == 'virtual'
        else: publish =  False

    if publish and has_element(node, ['argsstring']):
        publish = not get_text_for_element(node, 'argsstring').endswith('=0')

    if is_function: print_debug_info(node, ['kind', 'prot', 'static', 'virt'])
    else: print_debug_info(node, ['kind', 'prot', 'static'])
    if debug: print '  is_class:', is_class
    report(library, publish, symbol+'*')
    if is_function and node.attributes['virt'].value == 'virtual': report(library, publish, 'non-virtual?thunk?to?'+symbol+'*')

def parse_compound_defs(xmldoc):
    compounddefs = xmldoc.getElementsByTagName('compounddef')
    for node in compounddefs:
        kind = node.attributes['kind'].value

        if kind in ['page', 'file', 'example', 'union']: continue

        if kind in ['group']:
            for member in node.getElementsByTagName('memberdef') :
                parse_member_def(None, member, False)
            continue

        if kind in ['namespace']:
            symbol = concat_text_from_tags(node, ['compoundname'])
            for member in node.getElementsByTagName('memberdef') :
                parse_member_def(symbol, member, False)
            continue

        file = get_file_location(node)
        if debug: print '  from file:', file
        if '/examples/' in file or '/test/' in file or '[generated]' in file or '[STL]' in file:
            continue

        if has_element(node, ['templateparamlist']): continue

        library = mapped_physical_component(file)
        symbol = concat_text_from_tags(node, ['compoundname'])

        file_location = get_file_location(node)
        publish = is_file_publishable(file_location)

        if publish:
            if kind in ['class', 'struct']:
                prot =  node.attributes['prot'].value
                publish = prot != 'private'
                print_debug_info(node, ['kind', 'prot'])
                report(library, publish, 'vtable?for?' + symbol)
                report(library, publish, 'typeinfo?for?' + symbol)

        if publish:
            for member in node.getElementsByTagName('memberdef') :
                parse_member_def(symbol, member, kind in ['class', 'struct'])

if __name__ == "__main__":
    if len(argv) == 1 or '-h' in argv or '--help' in argv:
        print helptext
        exit()

    for arg in argv[1:]:
        try:
            if debug: print 'Processing:', arg
            xmldoc = minidom.parse(arg)
            parse_compound_defs(xmldoc)
        except Exception as error:
            print 'Error:', arg, error

    print_report()
